#include <cstdlib>
#include <stdexcept>

#include "SCADOPT.hpp"

template<typename CT, std::size_t DIM>
class TestCheckGradientTf : public SCADOPT::Problem< TestCheckGradientTf<CT,DIM>, CT, DIM > {
public:
    TestCheckGradientTf(ScaFES::Parameters const& cl,
                ScaFES::GridGlobal<DIM> const& gg,
            bool useLeapfrog,
            std::vector<std::string> const& nameDatafield,
            std::vector<int> const& stencilWidth,
            std::vector<bool> const& isKnownDf,
            std::vector<int> const& nLayers = std::vector<int>(),
            std::vector<bool> const& optimizeForDf = std::vector<bool>(),
            std::vector<CT> const& defaultValue = std::vector<CT>(),
            std::vector<ScaFES::WriteHowOften> const& writeToFile
                = std::vector<ScaFES::WriteHowOften>(),
            std::vector<bool> const& computeError = std::vector<bool>(),
            std::vector<CT> const& geomparamsInit = std::vector<CT>(),
            std::vector<CT> const& geomparamsSol = std::vector<CT>() )
            : SCADOPT::Problem<TestCheckGradientTf<CT, DIM>, CT, DIM>(
                     cl, gg, useLeapfrog, nameDatafield, stencilWidth,
                     isKnownDf, nLayers, optimizeForDf,
                     defaultValue, writeToFile,
                     computeError, geomparamsInit, geomparamsSol) { }

    template<typename TT>
    void evalInner(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                   ScaFES::Ntuple<int,DIM> const& idxNode,
                   int const& /*timestep*/) {
        vNew[0](idxNode) = 0.0;
        vNew[1](idxNode) = 0.0;
    }
    template<typename TT>
    void evalBorder(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                    ScaFES::Ntuple<int,DIM> const& idxNode,
                    int const& timestep) {
        this->evalInner(vNew, idxNode, timestep);
    }
    template<typename TT>
    void initInner(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                  std::vector<TT> const& vOld,
                  ScaFES::Ntuple<int,DIM> const& idxNode,
                  int const& /*timestep*/) {
        vNew[0](idxNode) = 2.0 * vOld[0]  + 3.0 * vOld[1];
        vNew[1](idxNode) = 5.0 * vOld[0]  + 7.0 * vOld[1];
    }
    template<typename TT>
    void initBorder(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                    std::vector<TT> const& vOld,
                    ScaFES::Ntuple<int,DIM> const& idxNode,
                    int const& timestep) {
        this->template initInner<TT>(vNew, vOld, idxNode, timestep);
    }
    template<typename TT>
    void updateInner(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                     std::vector< ScaFES::DataField<TT,DIM> > const& vOld,
                     ScaFES::Ntuple<int,DIM> const& idxNode,
                     int const& /*timestep*/) {
        vNew[0](idxNode) = 11.0 * vOld[0](idxNode) + 13.0 * vOld[1](idxNode);
    }
    template<typename TT>
    void updateBorder(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                      std::vector< ScaFES::DataField<TT,DIM>> const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& timestep) {
        this->template updateInner<TT>(vNew, vOld, idxNode, timestep);
    }
    template<typename TT>
    void updateInner2(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                     std::vector< ScaFES::DataField<TT,DIM> > const& vOld,
                     ScaFES::Ntuple<int,DIM> const& idxNode,
                     int const& /*timestep*/) {
        vNew[1](idxNode) = 17.0 * vOld[0](idxNode) + 19.0 * vOld[1](idxNode);
    }
    template<typename TT>
    void updateBorder2(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                      std::vector< ScaFES::DataField<TT,DIM>> const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& timestep) {
        this->template updateInner2<TT>(vNew, vOld, idxNode, timestep);
    }
    template<typename TT>
    void compTfInner(TT& vNew,
                     std::vector< ScaFES::DataField<TT,DIM> > const& vOld,
                     ScaFES::Ntuple<int,DIM> const& idxNode,
                     int const& /*timestep*/) {
        vNew = 23.0 * vOld[0](idxNode) + 27.0 * vOld[1](idxNode);
    }
    template<typename TT>
    void compTfBorder(TT& vNew,
        std::vector< ScaFES::DataField<TT,DIM> > const& vOld,
        ScaFES::Ntuple<int,DIM> const& idxNode, int const& timestep) {
        this->template compTfInner<TT>(vNew, vOld, idxNode, timestep);
    }
};
