#include <SCADOPT.hpp>
#include <TestCheckGradientTf.hpp>

const int DIM = 3;

int main(int argc, char *argv[]) {
    ScaFES::Parameters paramsCl(argc, argv);
    ScaFES::GridGlobal<DIM> gg(paramsCl);

    std::vector<std::string> nameDatafield(2);
    nameDatafield[0] = "U";
    nameDatafield[1] = "Y";
    std::vector<int> stencilWidth(2);
    stencilWidth[0] = 0;
    stencilWidth[1] = 0;
    std::vector<bool> isKnownDf(2);
    isKnownDf[0] = false;
    isKnownDf[1] = false;
    std::vector<int> nLayers(2);
    nLayers[0] = 0;
    nLayers[1] = 0;
    std::vector<bool> optimizeForDf(2);
    optimizeForDf[0] = false;
    optimizeForDf[1] = false;
    std::vector<double> defaultValue(2);
    defaultValue[0] = 0.0;
    defaultValue[1] = 0.0;
    std::vector<ScaFES::WriteHowOften> writeToFile(2);
    writeToFile[0] = ScaFES::WriteHowOften::NEVER;
    writeToFile[1] = ScaFES::WriteHowOften::AT_START;
    std::vector<bool> computeError(2);
    computeError[0] = false;
    computeError[1] = false;

    std::vector<double> geomparamsInit(2, 19); geomparamsInit[0] = 23;
    std::vector<double> geomparamsSol(2, 34); geomparamsSol[0] = 73;

    std::vector<bool> useLeapfrogScheme(2);
    useLeapfrogScheme[0] = false;
    useLeapfrogScheme[1] = true;
    for (std::size_t jj = 0; jj < useLeapfrogScheme.size(); ++jj) {
        TestCheckGradientTf<double, DIM> ppp(paramsCl, gg, 
                     useLeapfrogScheme.at(jj),
                     nameDatafield, stencilWidth,
                     isKnownDf, nLayers, optimizeForDf,
                     defaultValue, writeToFile,
                     computeError, geomparamsInit, geomparamsSol);
        ppp.compNewTfAndGradTf();

//         if (1 < paramsCl.nTimesteps()) {
//             throw std::runtime_error("Only 1 time step.");
//         }

        const double EPS = 2.2e-10;

        // Case: Tf is time-independent.
        // Multiply tf value of ONE timestep with number of timesteps.
        double tfNum = ppp.tf();
        double tfRef = 0.0;
        std::size_t rank = paramsCl.myWorld().rank();
        double tfTmp = 0.0;
        double vNew0 = 0.0;
        double vNew1 = 0.0;
        double vOld0 = 0.0;
        double vOld1 = 0.0;
        double vTmp0 = 0.0;
        double vTmp1 = 0.0;
        // Check tf.
        for (int ii = 0; ii < gg.partition(rank).nNodesSub().size(); ++ii) {
            vOld0 = 2 * geomparamsInit[0] + 3 * geomparamsInit[1];
            vOld1 = 5 * geomparamsInit[0] + 7 * geomparamsInit[1];
            for (int ii = 0; ii < paramsCl.nTimesteps(); ++ii) {
                vNew0 = 11 * vOld0 + 13 * vOld1;
                vNew1 = 1 * vOld1;
                if (1 == jj) {
                    vTmp0 = vOld0;
                    vOld0 = vNew0;
                    vNew0 = vTmp0;
                    vTmp1 = vOld1;
                    vOld1 = vNew1;
                    vNew1 = vTmp1;
                    vNew0 = 1 * vOld0;
                    vNew1 = 17 * vOld0 + 19 * vOld1;
                }
                vTmp0 = vOld0;
                vOld0 = vNew0;
                vNew0 = vTmp0;
                vTmp1 = vOld1;
                vOld1 = vNew1;
                vNew1 = vTmp1;
            }
            tfTmp = 23 * vOld0 + 27 * vOld1;
            tfRef += tfTmp;
        }
        std::cout << "tfNum = " << tfNum << std::endl;
        std::cout << "tfRef = " << tfRef << std::endl;
        if (EPS < ::fabs(tfRef - tfNum)) {
            throw std::runtime_error("Value of tf not correct!");
        }

        // Check gradient of tf.
        double gradTfTmp0 = 0.0;
        double gradTfTmp1 = 0.0;
        double vNew00 = 0.0;
        double vNew01 = 0.0;
        double vNew10 = 0.0;
        double vNew11 = 0.0;
        double vOld00 = 0.0;
        double vOld01 = 0.0;
        double vOld10 = 0.0;
        double vOld11 = 0.0;
        double vTmp00 = 0.0;
        double vTmp01 = 0.0;
        double vTmp10 = 0.0;
        double vTmp11 = 0.0;
        if (0 < paramsCl.enabledAdolc() && 0 < paramsCl.computeGradients()) {
            std::vector<double> gradTfNum = ppp.gradTf();
            std::vector<double> gradTfRef(2, 0.0);
            for (int ii = 0; ii < gg.partition(rank).nNodesSub().size(); ++ii) {
                vOld00 = 2;
                vOld01 = 3;
                vOld10 = 5;
                vOld11 = 7;
                for (int ii = 0; ii < paramsCl.nTimesteps(); ++ii) {
                    vNew00 = 11 * vOld00 + 13 * vOld10;
                    vNew01 = 11 * vOld01 + 13 * vOld11;
                    vNew10 = 1 * vOld10;
                    vNew11 =              1 * vOld11;
                    if (1 == jj) {
                        vTmp00 = vOld00;
                        vOld00 = vNew00;
                        vNew00 = vTmp00;
                        vTmp01 = vOld01;
                        vOld01 = vNew01;
                        vNew01 = vTmp01;
                        vTmp10 = vOld10;
                        vOld10 = vNew10;
                        vNew10 = vTmp10;
                        vTmp11 = vOld11;
                        vOld11 = vNew11;
                        vNew11 = vTmp11;
                        vNew00 = 1 * vOld00;
                        vNew01 = 1 * vOld01;
                        vNew10 = 17 * vOld00 + 19 * vOld10;
                        vNew11 = 17 * vOld01 + 19 * vOld11;
                    }
                    vTmp00 = vOld00;
                    vOld00 = vNew00;
                    vNew00 = vTmp00;
                    vTmp01 = vOld01;
                    vOld01 = vNew01;
                    vNew01 = vTmp01;
                    vTmp10 = vOld10;
                    vOld10 = vNew10;
                    vNew10 = vTmp10;
                    vTmp11 = vOld11;
                    vOld11 = vNew11;
                    vNew11 = vTmp11;
                }
                gradTfTmp0 = 23 * vOld00 + 27 * vOld10;
                gradTfTmp1 = 23 * vOld01 + 27 * vOld11;
                gradTfRef[0] += gradTfTmp0;
                gradTfRef[1] += gradTfTmp1;
            }
            std::cout << "gradTfRef[0] = " << gradTfRef[0] << std::endl;
            std::cout << "gradTfRef[1] = " << gradTfRef[1] << std::endl;
            std::cout << "gradTfNum[0] = " << gradTfNum[0] << std::endl;
            std::cout << "gradTfNum[1] = " << gradTfNum[1] << std::endl;
            for (std::size_t ii = 0; ii < gradTfNum.size(); ++ii) {
                if (EPS < ::fabs(gradTfRef[ii] - gradTfNum[ii])) {
                    throw std::runtime_error("Value of gradTf not correct!");
                }
            }
        }
    }

    return 0;
}

