#include <cstdlib>
#include <stdexcept>
#include <vector>
#include <cmath>

#include "SCADOPT.hpp"

template<typename CT, std::size_t DIM>
class TestInitOfOldAndNew
    : public SCADOPT::Problem< TestInitOfOldAndNew<CT,DIM>, CT, DIM > {
public:
    TestInitOfOldAndNew(ScaFES::Parameters const& cl,
                ScaFES::GridGlobal<DIM> const& gg,
            bool const& useLeapfrogScheme,
            std::vector<std::string> const& nameDatafield,
            std::vector<int> const& stencilWidth,
            std::vector<bool> const& isKnownDf,
            std::vector<int> const& nLayers = std::vector<int>(),
            std::vector<bool> const& optimizeForDf = std::vector<bool>(),
            std::vector<CT> const& defaultValue = std::vector<CT>(),
            std::vector<ScaFES::WriteHowOften> const& writeToFile
                = std::vector<ScaFES::WriteHowOften>(),
            std::vector<bool> const& computeError = std::vector<bool>(),
            std::vector<CT> const& geomparamsInit = std::vector<CT>(),
            std::vector<CT> const& geomparamsSol = std::vector<CT>() )
            : SCADOPT::Problem<TestInitOfOldAndNew<CT, DIM>, CT, DIM>(
                     cl, gg, useLeapfrogScheme, nameDatafield, stencilWidth,
                     isKnownDf, nLayers, optimizeForDf,
                     defaultValue, writeToFile,
                     computeError, geomparamsInit, geomparamsSol) { }

    template<typename TT>
    void evalInner(std::vector< ScaFES::DataField<TT,DIM> >& /*vNew*/,
                   ScaFES::Ntuple<int,DIM> const& /*idxNode*/,
                   int const& /*timestep*/) { }
    template<typename TT>
    void evalBorder(std::vector< ScaFES::DataField<TT,DIM> >& /*vNew*/,
                    ScaFES::Ntuple<int,DIM> const& /*idxNode*/,
                    int const& /*timestep*/) { }
    template<typename TT>
    void initInner(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                  std::vector<TT> const& /*vOld*/,
                  ScaFES::Ntuple<int,DIM> const& idxNode,
                  int const& /*timestep*/) {
        const double PI = 3.14159265358979323846;
        for (std::size_t ii = 0; ii < vNew.size(); ++ii) {
            vNew[ii](idxNode) = PI;
        }
    }
    template<typename TT>
    void initBorder(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                    std::vector<TT> const& vOld,
                    ScaFES::Ntuple<int,DIM> const& idxNode,
                    int const& timestep) {
        this->template initInner<TT>(vNew, vOld, idxNode, timestep);
    }
    template<typename TT>
    void updateInner(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                     std::vector< ScaFES::DataField<TT,DIM> > const& vOld,
                     ScaFES::Ntuple<int,DIM> const& idxNode,
                     int const& /*timestep*/) {
        for (std::size_t ii = 0; ii < vNew.size(); ++ii) {
            if (vNew[ii](idxNode) != vOld[ii](idxNode)) {
                std::stringstream what;
                what << "Difference between old and new value @ "
                      << idxNode << ": old= "
                      << std::setw(8) << std::setprecision(16)
                      << vOld[ii](idxNode)
                      << " <-> new= "
                      << std::setw(8) << std::setprecision(16)
                      << vNew[ii](idxNode);
                throw std::runtime_error(what.str());
            }
        }
    }
    template<typename TT>
    void updateBorder(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                      std::vector< ScaFES::DataField<TT,DIM>> const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& timestep) {
        this->template updateInner<TT>(vNew, vOld, idxNode, timestep);
    }
    template<typename TT>
    void updateInner2(std::vector< ScaFES::DataField<TT,DIM> >& /*vNew*/,
                     std::vector< ScaFES::DataField<TT,DIM> > const& /*vOld*/,
                     ScaFES::Ntuple<int,DIM> const& /*idxNode*/,
                     int const& /*timestep*/) { }
    template<typename TT>
    void updateBorder2(std::vector< ScaFES::DataField<TT,DIM> >& vNew,
                      std::vector< ScaFES::DataField<TT,DIM>> const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& timestep) {
        this->template updateInner2<TT>(vNew, vOld, idxNode, timestep);
    }
    template<typename TT>
    void compTfInner(TT& /*vNew*/,
                     std::vector< ScaFES::DataField<TT,DIM> > const& /*vOld*/,
                     ScaFES::Ntuple<int,DIM> const& /*idxNode*/,
                     int const& /*timestep*/) { }
    template<typename TT>
    void compTfBorder(TT& vNew,
        std::vector< ScaFES::DataField<TT,DIM> > const& vOld,
        ScaFES::Ntuple<int,DIM> const& idxNode, int const& timestep) {
        this->template compTfInner<TT>(vNew, vOld, idxNode, timestep);
    }
};
