#include <cstdlib>
#include <stdexcept>

#include <SCADOPT.hpp>

#ifdef _OPENMP
#include <omp.h>
#endif

template<typename CT, std::size_t DIM = 3>
class TestLeapfrogNotUsedWhenNotDesired
    : public SCADOPT::Problem< TestLeapfrogNotUsedWhenNotDesired<CT,DIM>, CT, DIM> {
private:
    typedef TestLeapfrogNotUsedWhenNotDesired<CT,DIM> Self;
    typedef SCADOPT::Problem< Self, CT, DIM> Super;

    typedef ScaFES::Ntuple<int, DIM> Index;

    bool touched;

public:
    TestLeapfrogNotUsedWhenNotDesired(ScaFES::Parameters const& cl,
                                      ScaFES::GridGlobal<DIM> const& gg,
            bool useLeapfrog,
            std::vector<std::string> const& nameDatafield,
            std::vector<int> const& stencilWidth,
            std::vector<bool> const& isKnownDf,
            std::vector<int> const& nLayers = std::vector<int>(),
            std::vector<bool> const& optimizeForDf = std::vector<bool>(),
            std::vector<CT> const& defaultValue = std::vector<CT>(),
            std::vector<ScaFES::WriteHowOften> const& writeToFile
                = std::vector<ScaFES::WriteHowOften>(),
            std::vector<bool> const& computeError = std::vector<bool>(),
            std::vector<CT> const& geomparamsInit = std::vector<CT>(),
            std::vector<CT> const& geomparamsSol = std::vector<CT>() )
            : SCADOPT::Problem<TestLeapfrogNotUsedWhenNotDesired<CT, DIM>, CT, DIM>(
                     cl, gg, useLeapfrog, nameDatafield, stencilWidth,
                     isKnownDf, nLayers, optimizeForDf,
                     defaultValue, writeToFile,
                     computeError, geomparamsInit, geomparamsSol)
        , touched(false) { }

    template<typename TT>
    void initInner(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*vOld*/,
            Index const& /*idxNode*/,
            int const& /*timestep*/) { }
    template<typename TT>
    void initBorder(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*vOld*/,
            Index const& /*idxNode*/,
            int const& /*timestep*/) { }
    template<typename TT>
    void initInner(std::vector< ScaFES::DataField<TT, DIM> >& /*dfs*/,
            std::vector<TT> const& /*p*/,
            Index const& /*idx*/,
            int const& /*timestep*/) { }
    template<typename TT>
    void initBorder(std::vector< ScaFES::DataField<TT, DIM> >& /*dfs*/,
            std::vector<TT> const& /*p*/,
            Index const& /*idx*/,
            int const& /*timestep*/) { }

    template<typename TT>
    void updateInner(std::vector< ScaFES::DataField<TT, DIM> >& /*n*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*o*/,
            Index const& /*idx*/,
            int const& timestep) {
        if (1 < timestep)
            if (this->touched)
                throw std::runtime_error("Leapfrog step done, but leapfrog not enabled.");
    }

    template<typename TT>
    void updateBorder(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
            std::vector< ScaFES::DataField<TT, DIM>>const& /*vOld*/,
            Index const& /*idx*/,
            int const& /*timestep*/) {}

    template<typename TT>
    void updateInner2(std::vector< ScaFES::DataField<TT, DIM> >& /*n*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*o*/,
            Index const& /*idx*/,
            int const& /*timestep*/) {
        this->touched=true;
    }

    template<typename TT>
    void updateBorder2(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
            std::vector< ScaFES::DataField<TT, DIM>>const& /*vOld*/,
            Index const& /*idx*/,
            int const& /*timestep*/) {}

    template<typename TT>
    void compTfInner(TT& /*tf*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*vOld*/,
            Index const& /*idxNode*/,
            int const& /*timestep*/) { }
    template<typename TT>
    void compTfBorder(TT& /*tf*/,
            std::vector< ScaFES::DataField<TT, DIM> > const& /*vOld*/,
            Index const& /*idxNode*/,
            int const& /*timestep*/) { }

    template<typename TT>
    void evalInner(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
        Index const& /*idxNode*/,
        int const& /*timestep*/) { }
    template<typename TT>
    void evalBorder(std::vector< ScaFES::DataField<TT, DIM> >& /*vNew*/,
        Index const& /*idxNode*/,
        int const& /*timestep*/) { }
};
